<?php namespace SokoWeb\Database\Mysql;

use SokoWeb\Interfaces\Database\IConnection;
use SokoWeb\Interfaces\Database\IResultSet;
use SokoWeb\Interfaces\Database\IStatement;
use mysqli;

class Connection implements IConnection
{
    private string $host;
    private string $user;
    private string $password;
    private string $db;
    private int $port;
    private string $socket;
    private ?mysqli $connection = null;

    public function __construct(string $host, string $user, string $password, string $db, int $port = -1, string $socket = null)
    {
        $this->host = $host;
        $this->user = $user;
        $this->password = $password;
        $this->db = $db;
        $this->port = $port < 0 ? (int) ini_get('mysqli.default_port') : $port;
        $this->socket = $socket === null ? (string) ini_get('mysqli.default_socket') : $socket;
    }

    public function __destruct()
    {
        if ($this->connection === null) {
            return;
        }
        $this->connection->close();
    }

    public function startTransaction(): void
    {
        $this->getConnection()->autocommit(false);
    }

    public function commit(): void
    {
        $this->getConnection()->commit();
        $this->getConnection()->autocommit(true);
    }

    public function rollback(): void
    {
        $this->getConnection()->rollback();
        $this->getConnection()->autocommit(true);
    }

    public function query(string $query): ?IResultSet
    {
        $result = $this->getConnection()->query($query);

        if ($result !== true) {
            return new ResultSet($result);
        }

        return null;
    }

    public function multiQuery(string $query): array
    {
        $this->getConnection()->multi_query($query);

        $ret = [];
        do {
            if ($result = $this->getConnection()->store_result()) {
                $ret[] = new ResultSet($result);
            } else {
                $ret[] = null;
            }

            $this->getConnection()->more_results();
        } while ($this->getConnection()->next_result());

        return $ret;
    }

    public function prepare(string $query): IStatement
    {
        $stmt = $this->getConnection()->prepare($query);

        return new Statement($stmt);
    }

    public function lastId(): int
    {
        return $this->getConnection()->insert_id;
    }

    public function getAffectedRows(): int
    {
        return $this->getConnection()->affected_rows;
    }

    private function getConnection(): mysqli
    {
        if ($this->connection === null) {
            $this->createConnection();
        }
        return $this->connection;
    }

    private function createConnection(): void
    {
        mysqli_report(MYSQLI_REPORT_ERROR | MYSQLI_REPORT_STRICT);
        $this->connection = new mysqli($this->host, $this->user, $this->password, $this->db, $this->port, $this->socket);
        $this->connection->set_charset('utf8mb4');
    }
}
